--[[

   bossFight.lua
   Handles the JRPG end boss of Because it's fun, Fay.
	
   Copyright 2008 Oxeye Game Studio. http://www.oxeyegames.com
	
   File contributors:
      Jens Bergensten
      Daniel Brynolf

   This game was written for the TIGSource Procedural Generation
   competition in May 2008. You may use this Lua source code in
   any way you please, we don't mind. However, the game's sound
   effects and artwork belong to us, so please ask for permission
   before you do anything funny =)

]]

dofile("shellsort.lua")

local catSprites = {}
local partSprites= {}
local circleSprite = nil
local backgroundSprite, bottomSprite, victorySprite = nil, nil, nil
local guiSprites = {}
local caterpillar = {
	segments = {},
	mutations = {},
	isMutating = false,	hasMutated = false,
	isRetreating = false, hasRetreated = false,
	goingLeft = true,
	goingFlag = true,
	reachedSquare = false, errorTimer = 0,
	wingRotation = 0,
	bodyRotation = 0,
	velocity = 1200,
	slowingDown = false,
	lifeTimer = 0,
	soundTimer = 0,
	headShakeTimer = 0
}
local introTimer = 0 -- 8
local viewX = 0
local fightState = 0 -- 0 monster entry, 1 player powers, 2 fight, 3 monster reappears, 4 final fight
local roundState = 0 -- 0 player input, 1 executing orders
local activePlayer = 0
local gameOver, extroTimer, isVictory = false, 8, false
local roundDelays = { firstWait = 1, actionWait = 1, resultWait = 1, extraEvent = false }
local mutationTypes = {
	{	-- crab
		actionType = 1,
		attack = 35, magic = 10, defense = 30, agility = 10,
		startHealth = 300,
		r = 255, g = 170, b = 0
	},
	{	-- healer
		actionType = 2,
		attack = 10, magic = 40, defense = 20, agility = 40,
		startHealth = 150,
		r = 0, g = 170, b = 255
	},
	{	-- vampire
		actionType = 3,
		attack = 10, magic = 45, defense = 15, agility = 40,
		startHealth = 200,
		r = 160, g = 0, b = 64
	},
	{	-- clockwork
		actionType = 4,
		attack = 10, magic = 40, defense = 20, agility = 30,
		startHealth = 200,
		r = 190, g = 190, b = 190
	},
	{
		-- head
		actionType = 5,
		attack = 50, magic = 70, defense = 10, agility = 40,
		startHealth = 1000,
		r = 255, g = 0, b = 0
	}
}

local PY = 668
local PX1 = 350
local PX2 = 674
local playerPowers = {}
local textFloaters = {}
local powerTexts = { "ATK UP", "DEF UP", "MGK UP", "LIFE UP", "MANA UP", "AGI UP" }
local statUpdate = 0
-- spell types, 1 single party, 2 all party, 3 single enemy, 4 all enemy
local spells = {
	lak = {    name = "Lak", type = 1, agiBonus = 20, manaCost = 7, actionMethod = nil, testMethod = nil, sound = "lak.wav", timer = 0.4 },
	dulak = {  name = "DuLak", type = 2, agiBonus = 10, manaCost = 10, actionMethod = nil, testMethod = nil, sound = "lak.wav", timer = 0.4 },
	uppsta = { name = "Uppsta", type = 1, agiBonus = 100, manaCost = 13, actionMethod = nil, testMethod = nil, sound = "uppsta.wav", timer = 1.0 },
	skold = {  name = "Skold", type = 2, agiBonus = 0, manaCost = 9, actionMethod = nil, testMethod = nil, sound = "skold.wav", timer = 1.8 },
	hast = {   name = "Hast", type = 2, agiBonus = 0, manaCost = 9, actionMethod = nil, testMethod = nil, sound = "hast.wav", timer = 1.8 },
	brann = {  name = "Brann", type = 3, agiBonus = 0, manaCost = 4, actionMethod = nil, testMethod = nil, sound = "brann.wav", timer = 0.1 },
	blixt = {  name = "Blixt", type = 4, agiBonus = -10, manaCost = 9, actionMethod = nil, testMethod = nil, sound = "blixt.wav", timer = 0.1 },

	-- item "spells"
	dryer = { name = "Hair Dryer", type = 2, agiBonus = 20, manaCost = 0, actionMethod = nil, testMethod = nil, sound = "dryer.wav", timer = 0.5 },
	syringe = { name = "Syringe", type = 1, agiBonus = 20, manaCost = 0, actionMethod = nil, testMethod = nil, sound = "syringe.wav", timer = 0.4 }
}
local itemTypes = {
	dryer = { sprite = 3, spell = spells.dryer },
	oxeye = { sprite = 6, spell = nil },
	syringe = { sprite = 12, spell = spells.syringe }
}
local lowestPossibleMana = 4
local dryerCount = 0
local playerCharacters = {
	{
		-- STATS
		-- visible, actual, visible max, actual max
		health = { 100, 100, 100, 100 }, healthPeak = 260,
		mana = { 50, 50, 50, 50 }, manaPeak = 150,
		attack = 40, attackPeak = 50,
		defense = 25, defensePeak = 35,
		magic = 20, magicPeak = 30,
		agility = 30, agilityPeak = 50,
		isDefending = false, isShielded = 0, isHasted = 0,

		-- GUI
		screenX = PX1,
		done = false,
		state = 0, -- 0 main, 1 attack, 2 defense, 3 magic, 4 item, 5 spell target, 6 item target
		left = false, up = false, right = false, down = false,
		prevLeft = false, prevUp = false, prevRight = false, prevDown = false,
		selectionIndex = 0, subIndex = 0,
		selectionAction = 0,
		spells = { spells.lak, spells.uppsta, spells.skold, spells.hast, spells.brann },
		items = { },
		
		-- sprites
		deathSprite = -1, aliveSprites = {},
		deathSpriteOffsets = {x = -50, y = -10 },
		aliveSpriteOffsets = { { x = 4, y = 22, angle = 0 }, { x = -2, y = 22, angle = 0 }, { x = 5, y = -17, angle = 0 }, { x = 0, y = 0, angle = 0 }, { x = -10, y = -17, angle = 0 }, { x = 0, y = -27, angle = 0 } }

	},
	{
		-- STATS
		-- visible, actual, visible max, actual max
		health = { 100, 100, 100, 100 }, healthPeak = 180,
		mana = { 50, 50, 50, 50 }, manaPeak = 300,
		attack = 20, attackPeak = 30,
		defense = 20, defensePeak = 30,
		magic = 45, magicPeak = 80,
		agility = 25, agilityPeak = 30,
		isDefending = false, isShielded = 0, isHasted = 0,

		-- GUI
		screenX = PX2,
		done = false,
		state = 0, -- 0 main, 1 attack, 2 defense, 3 magic, 4 item, 5 spell target, 6 item target
		left = false, up = false, right = false, down = false,
		prevLeft = false, prevUp = false, prevRight = false, prevDown = false,
		selectionIndex = 0, subIndex = 0,
		selectionAction = 0,
		spells = { spells.lak, spells.dulak, spells.uppsta, spells.brann, spells.blixt },
		items = { },
		
		-- sprites
		deathSprite = -1, aliveSprites = {},
		deathSpriteOffsets = {x = -50, y = -25 },
		aliveSpriteOffsets = {
			{ x = -15, y = -18 + 30, angle = 0 }, { x = -20, y = -16 + 30, angle = -math.pi * .15 },
			{ x = 15, y = -18 + 30, angle = 0 }, { x = 20, y = -16 + 30, angle = math.pi * .15 },
			{ x = 15, y = 25 + 30, angle = -math.pi * .30 }, { x = -15, y = 25 + 30, angle = math.pi * .30 },
			{ x = 20, y = -15 + 30, angle = math.pi * .05 }, { x = -20, y = -15 + 30, angle = -math.pi * .05 },
			{ x = 0, y = 30, angle = 0 },
			{ x = 0, y = -16 + 30, angle = 0 } }
	}
}
local cornAnimTimer = 0
local roundActions = {}


local function renderStats(stats, x, y)

	daisy.renderRectangle(x, y, 200, 90, 140, 0, 0, 0)

	local healthColor = { 255, 255, 255 }
	if stats.health[2] < stats.health[4] / 4 then
		healthColor[1] = 160
		healthColor[2] = 0
		healthColor[3] = 0
	elseif stats.health[2] < stats.health[4] / 2 then
		healthColor[1] = 255
		healthColor[2] = 230
		healthColor[3] = 0
	end
	daisy.renderText("Life", x + 10, y + 10, 0, "gameFont.fnt", 255, healthColor[1], healthColor[2], healthColor[3])
	daisy.renderText(stats.health[1] .. " / " .. stats.health[3], x + 190, y + 10, 2, "gameFont.fnt", 255, healthColor[1], healthColor[2], healthColor[3])

	daisy.renderText("Mana", x + 10, y + 30, 0, "gameFont.fnt", 255, 255, 255, 255)
	daisy.renderText(stats.mana[1] .. " / " .. stats.mana[3], x + 190, y + 30, 2, "gameFont.fnt", 255, 255, 255, 255)

	daisy.renderText("ATK " .. stats.attack, x + 90, y + 50, 2, "gameFont.fnt", 255, 255, 255, 255)
	if stats.isShielded > 0 then
		daisy.renderText("DEF " .. stats.defense, x + 165, y + 50, 2, "gameFont.fnt", 255, 128, 255, 255)
	else
		daisy.renderText("DEF " .. stats.defense, x + 165, y + 50, 2, "gameFont.fnt", 255, 255, 255, 255)
	end

	daisy.renderText("MGK " .. stats.magic, x + 90, y + 70, 2, "gameFont.fnt", 255, 255, 255, 255)
	if stats.isHasted > 0 then
		daisy.renderText("AGI " .. stats.agility, x + 165, y + 70, 2, "gameFont.fnt", 255, 128, 255, 255)
	else
		daisy.renderText("AGI " .. stats.agility, x + 165, y + 70, 2, "gameFont.fnt", 255, 255, 255, 255)
	end

	--daisy.renderText("ATK " .. stats.attack .. "   DEF " .. stats.defense, x + 100, y + 50, 1, "gameFont.fnt", 255, 255, 255, 255)
	--daisy.renderText("MGK " .. stats.magic .. "   AGI " .. stats.agility, x + 100, y + 70, 1, "gameFont.fnt", 255, 255, 255, 255)

end

local function renderPlayerSprites(stats, x, y)
	if stats.health[2] <= 0 then
		daisy.renderSpriteState(stats.deathSprite, x + stats.deathSpriteOffsets.x, y + stats.deathSpriteOffsets.y, 2)
	else
		for i=1,#stats.aliveSprites do
			daisy.renderSpriteState(stats.aliveSprites[i], x + stats.aliveSpriteOffsets[i].x, y + stats.aliveSpriteOffsets[i].y, 2, stats.aliveSpriteOffsets[i].angle)
		end
	end
end

local function renderplayerCharacters()
	renderStats(playerCharacters[1], 10, PY)
	renderStats(playerCharacters[2], 1024 - 10 - 200, PY)

	if dryerCount > 0 then
		daisy.renderSpriteState(circleSprite, PX1, PY, 2.5, 0, 128)
		daisy.renderSpriteState(circleSprite, PX2, PY, 2.5, 0, 128)
	end
	renderPlayerSprites(playerCharacters[1], PX1, PY)
	renderPlayerSprites(playerCharacters[2], PX2, PY)
end

local function updateplayerCharacters(time)

	statUpdate = statUpdate - time
	if statUpdate <= 0 then
		statUpdate = statUpdate + .05

		for i=1,2 do
			if playerCharacters[i].health[1] < playerCharacters[i].health[2] then
				playerCharacters[i].health[1] = playerCharacters[i].health[1] + 1
			elseif playerCharacters[i].health[1] > playerCharacters[i].health[2] then
				playerCharacters[i].health[1] = playerCharacters[i].health[1] - 1
			end
			if playerCharacters[i].health[3] < playerCharacters[i].health[4] then
				playerCharacters[i].health[3] = playerCharacters[i].health[3] + 1
			elseif playerCharacters[i].health[3] > playerCharacters[i].health[4] then
				playerCharacters[i].health[3] = playerCharacters[i].health[3] - 1
			end
			if playerCharacters[i].mana[1] < playerCharacters[i].mana[2] then
				playerCharacters[i].mana[1] = playerCharacters[i].mana[1] + 1
			elseif playerCharacters[i].mana[1] > playerCharacters[i].mana[2] then
				playerCharacters[i].mana[1] = playerCharacters[i].mana[1] - 1
			end
			if playerCharacters[i].mana[3] < playerCharacters[i].mana[4] then
				playerCharacters[i].mana[3] = playerCharacters[i].mana[3] + 1
			elseif playerCharacters[i].mana[3] > playerCharacters[i].mana[4] then
				playerCharacters[i].mana[3] = playerCharacters[i].mana[3] - 1
			end
		end
	end
end

local function addTextFloater(text, x, y)
	table.insert(textFloaters, {text=text, x=x, y=y, tx=x, ty=y-100, alpha=255})
end

local function updateTextFloaters(time)
	for i,text in pairs(textFloaters) do
		text.x = text.x + (text.tx - text.x) * time * 2
		text.y = text.y + (text.ty - text.y) * time * 2
		text.alpha = text.alpha + -text.alpha * time * 1

		if text.alpha < 1 then
			table.remove(textFloaters, i)
		end
	end
end

local function renderTextFloaters()
	for i,text in pairs(textFloaters) do
		daisy.renderText(text.text, text.x, text.y, 1, "gameFont.fnt", text.alpha, 255, 255, 255)
	end
end

local powerSfxCount = 0
local function applyPower(power)
	addTextFloater(powerTexts[power.type], power.x, power.y - 20)
	daisy.playSound("powerup.wav", 1.0, 0, 1.0 + powerSfxCount * .01)
	powerSfxCount = powerSfxCount + 1

	local stats = playerCharacters[1]
	if power.player == 1 then
		stats = playerCharacters[2]
	end

	if power.type == 1 then
		stats.attack = stats.attack + 1
		if stats.attack < stats.attackPeak then
			stats.attack = stats.attack + 2
		end
	elseif power.type == 2 then
		stats.defense = stats.defense + 1
		if stats.defense < stats.defensePeak then
			stats.defense = stats.defense + 2
		end
	elseif power.type == 3 then
		stats.magic = stats.magic + 1
		if stats.magic < stats.magicPeak then
			stats.magic = stats.magic + 2
		end
	elseif power.type == 4 then
		stats.health[2] = stats.health[2] + 7
		stats.health[4] = stats.health[4] + 7
		if stats.health[4] < stats.healthPeak then
			stats.health[2] = stats.health[2] + 19
			stats.health[4] = stats.health[4] + 19
		end
	elseif power.type == 5 then
		stats.mana[2] = stats.mana[2] + 6
		stats.mana[4] = stats.mana[4] + 6
		if stats.mana[4] < stats.manaPeak then
			stats.mana[2] = stats.mana[2] + 11
			stats.mana[4] = stats.mana[4] + 11
		end
	elseif power.type == 6 then
		stats.agility = stats.agility + 1
		if stats.agility < stats.agilityPeak then
			stats.agility = stats.agility + 1
		end
	end
end

local function presentPlayerPowers()
	fightState = 1

	local timeOff = 1

	-- create player powers
	for i=1,#(playerPowerups.fay) do
		local spriteType = playerPowerups.fay[i]
		local effectType = playerPowerups.getPowerupType(spriteType)
		if effectType >= 1 then
			local x = PX1
			local player = 0
			local angle = math.pi * 1.05 + math.random() * math.pi * 0.9
			local tx = x + math.cos(angle) * 150
			local ty = 668 + math.sin(angle) * 150
			table.insert(playerPowers, { x=x, y=668, type=effectType, sprite=powerupSprites[spriteType], scale=0, tScale=1, tx=tx, ty=ty, alpha=0, tAlpha=255, delay=timeOff * .05, fallbackDelay=3+timeOff*.5, player=player })
			timeOff = timeOff + 1
		else
			-- add as item instead
			local item = itemTypes.dryer
			if spriteType == 6 then
				item = itemTypes.oxeye
			elseif spriteType == 12 then
				item = itemTypes.syringe
			end
			table.insert(playerCharacters[1].items, item)
		end
	end
	timeOff = 1
	for i=1,#(playerPowerups.corn) do
		local spriteType = playerPowerups.corn[i]
		local effectType = playerPowerups.getPowerupType(spriteType)
		if effectType >= 1 then
			local x = PX2
			local player = 1
			local angle = math.pi * 1.05 + math.random() * math.pi * 0.9
			local tx = x + math.cos(angle) * 150
			local ty = 668 + math.sin(angle) * 150
			table.insert(playerPowers, { x=x, y=668, type=effectType, sprite=powerupSprites[spriteType], scale=0, tScale=1, tx=tx, ty=ty, alpha=0, tAlpha=255, delay=.5 + timeOff * .05, fallbackDelay=3.4+timeOff*.5, player=player })
			timeOff = timeOff + 1
		else
			local item = itemTypes.dryer
			if spriteType == 6 then
				item = itemTypes.oxeye
			elseif spriteType == 12 then
				item = itemTypes.syringe
			end
			table.insert(playerCharacters[2].items, item)
		end
	end
end

local function updatePlayerPowers(time)

	if fightState == 1 then
		for i,power in pairs(playerPowers) do

			power.delay = power.delay - time
			power.fallbackDelay = power.fallbackDelay - time
			if power.fallbackDelay <= 0 then
				if power.player == 0 then
					power.tx = PX1
				else
					power.tx = PX2
				end
				power.ty = 668
			end
			if power.delay <= 0 then
				power.x = power.x + (power.tx - power.x) * time * 4
				power.y = power.y + (power.ty - power.y) * time * 4
				power.scale = power.scale + (power.tScale - power.scale) * time * 2
				power.alpha = power.alpha + (power.tAlpha - power.alpha) * time * 2

				if power.fallbackDelay <= 0 and math.abs(power.x - power.tx) < 3 and math.abs(power.y - power.ty) < 3 then
					applyPower(power)
					table.remove(playerPowers, i)
				end
			end
		end

		if #playerPowers == 0 and #textFloaters == 0 then
			fightState = 2
		end
	end
end

local function renderPlayerPowers()
	local glow = powerupSprites.glow
	for i,power in pairs(playerPowers) do

		if power.alpha > 0 and power.scale > 0 then
			daisy.renderSpriteState(glow, power.x, power.y, power.scale, 0, power.alpha)
			daisy.renderSpriteState(power.sprite, power.x, power.y, power.scale, 0, power.alpha)
		end

	end
end

--   SSSSSS
-- SS      SS
--   SS
--     SS
--       SS
-- SS      SS
--   SSSSSS   E G M E N T S
local function addSegment(type)

	local segs = caterpillar.segments
	segs[#segs + 1] = { type=type, angle=math.pi * 1.5, x=math.random(300,724), y=1600 + #segs * 70, mutated = false, vx=math.random(-100,100), vy=0, fallTimer=0 }
end

local function addMutation(mutType, segment, segIndex)
	table.insert(caterpillar.mutations, {
		seg=segment,
		segIndex=segIndex,
		timer=.5 + math.random() * 1.5,
		type=mutType,
		health=mutType.startHealth,
		isDefending = false,
		animTimer = 0,
		actionTimer = 0,
		randValue = math.random(),
		animState = "birth"
		})
end

local function getAngle(dx,dy)
	local wantedAngle = 0
	if dx == 0 then
		if dy < 0 then
			wantedAngle = math.pi * 1.5
		else
			wantedAngle = math.pi * 0.5
		end
	else
		wantedAngle = math.atan(dy / dx)
		if dx < 0 then
			wantedAngle = wantedAngle + math.pi
		end
		if wantedAngle < 0 then
			wantedAngle = wantedAngle + 2 * math.pi
		elseif wantedAngle >= 2 * math.pi then
			wantedAngle = wantedAngle - 2 * math.pi
		end
	end
	return wantedAngle
end

local function updateSegments(time)

	caterpillar.wingRotation = caterpillar.wingRotation + time * 40
	caterpillar.bodyRotation = caterpillar.bodyRotation + time * 3.5
	caterpillar.lifeTimer = caterpillar.lifeTimer + time
	caterpillar.soundTimer = caterpillar.soundTimer + time
	caterpillar.headShakeTimer = caterpillar.headShakeTimer - time

	if caterpillar.soundTimer > .05 and not caterpillar.slowingDown then
		local index = math.random(1,2)
		local pitch = 0.9 + math.random() * 0.25
		caterpillar.soundTimer = 0
		daisy.playSound("step" .. index .. ".wav", 1.0, 0.0, pitch)
	end

	local segs = caterpillar.segments

	if fightState < 3 then

		if caterpillar.hasMutated then
			for index,mut in pairs(caterpillar.mutations) do
				if mut.actionTimer > 0 then
					mut.actionTimer = mut.actionTimer - time
					mut.animTimer = mut.animTimer + time * 2
				else
					mut.animTimer = mut.animTimer + time * .25
				end
			end
		end

		-- set the first segment
		--segs[1].x, segs[1].y = daisy.getMousePosition()
		if caterpillar.isMutating and not caterpillar.hasMutated then

			local found = false
			for i=1,#caterpillar.mutations do
				local mut = caterpillar.mutations[i]
				if mut.timer > 0 then
					mut.timer = mut.timer - time

					if mut.timer <= 0 then
						mut.seg.mutated = true
						caterpillar.headShakeTimer = .25
					end
					found = true
					break
				end
			end

			if not found and caterpillar.headShakeTimer <= 0 then
				-- time to show the player's powers
				caterpillar.hasMutated = true
				presentPlayerPowers()
			end

		elseif not caterpillar.hasMutated then
			if segs[1].x < 300 or segs[1].x > 724 or segs[1].y < 200 or segs[1].y > 368 then
				if caterpillar.reachedSquare then
					if caterpillar.goingFlag then
						if math.random() > .3 then
							caterpillar.goingLeft = not caterpillar.goingLeft
						end
						caterpillar.goingFlag = false
					end
					if caterpillar.goingLeft then
						segs[1].angle = segs[1].angle + time * (caterpillar.velocity / 200)
					else
						segs[1].angle = segs[1].angle - time * (caterpillar.velocity / 200)
					end
					if segs[1].angle < 0 then
						segs[1].angle = segs[1].angle + math.pi * 2
					elseif segs[1].angle > math.pi * 2 then
						segs[1].angle = segs[1].angle - math.pi * 2
					end

					-- hack to avoid stuck caterpillars
					caterpillar.errorTimer = caterpillar.errorTimer + time
					if caterpillar.errorTimer > 10 then
						if segs[1].x < 300 then
							segs[1].x = 290
						elseif segs[1].x > 724 then
							segs[1].x = 714
						end
						if segs[1].y < 200 then
							segs[1].y = 190
						elseif segs[1].y > 368 then
							segs[1].y = 358
						end
					end
				end
			else
				caterpillar.goingFlag = true
				caterpillar.reachedSquare = true
				caterpillar.errorTimer = 0

				if caterpillar.lifeTimer > 6 and not caterpillar.slowingDown then

					-- scan for mutation targets
					local headX = segs[1].x - 512
					for i=3,#segs-2,2 do

						if math.random() > (.25 + #caterpillar.mutations * .05) then
							local x,y = segs[i].x, segs[i].y
							if (x - headX) > 100 and (x - headX) < 924 and y > 100 and y < 500 then

								-- may mutate if not covered by other body parts
								local found = false
								for j=1,i-6,2 do
									local dx = segs[j].x - x
									local dy = segs[j].y - y

									if ((dx*dx) + (dy*dy)) < (120*120) then
										found = true
										break
									end
								end
								if not found then
									-- M U T A T I O N
									-- M U T A T I O N
									-- M U T A T I O N
									local type = mutationTypes[math.random(1, 4)]
									addMutation(type, segs[i], i)
									-- M U T A T I O N
									-- M U T A T I O N
									-- M U T A T I O N
								end
							end
						end

					end

					if #(caterpillar.mutations) >= 3 then
						caterpillar.slowingDown = true
						daisy.playMusic("darkForce.mp3", 0.75, true)

						-- boost the enemy in case there are few mutations
						local boost = 0
						if #(caterpillar.mutations) == 3 then
							boost = 10
						elseif #(caterpillar.mutations) == 4 then
							boost = 5
						end
						if boost > 0 then
							for typeIndex,mutationType in pairs(mutationTypes) do
								mutationType.defense = mutationType.defense + boost
							end
						end
					else
						caterpillar.lifeTimer = caterpillar.lifeTimer - 1
						caterpillar.mutations = {}
					end
				end
			end
			if caterpillar.slowingDown then
				caterpillar.velocity = caterpillar.velocity - caterpillar.velocity * time * 10

				if caterpillar.velocity < .5 then
					caterpillar.isMutating = true
				end
			end
		end

		if not caterpillar.isMutating then
			segs[1].x = segs[1].x + math.cos(segs[1].angle) * time * caterpillar.velocity
			segs[1].y = segs[1].y + math.sin(segs[1].angle) * time * caterpillar.velocity

			-- update following segments
			for i=3,#segs,2 do
				local cat = segs[i]
				local prevCat = segs[i-2]
				local dx,dy = prevCat.x - cat.x, prevCat.y - cat.y

				local lookAngle = getAngle(dx,dy)
				local diffAngle = lookAngle - prevCat.angle
				if diffAngle < 0 then
					diffAngle = diffAngle + math.pi * 2
				end
				if (diffAngle > math.pi * .25 and diffAngle < math.pi) or (diffAngle > 2.25 * math.pi and diffAngle < math.pi * 3) then
					lookAngle = prevCat.angle + math.pi * .25
					if lookAngle > math.pi * 2 then
						lookAngle = lookAngle - math.pi * 2
					end
				elseif (diffAngle < math.pi * 1.75 and diffAngle > math.pi) or (diffAngle < -.25 * math.pi and diffAngle > -math.pi) then
					lookAngle = prevCat.angle - math.pi * .25
					if lookAngle < 0 then
						lookAngle = lookAngle + math.pi * 2
					end
				end

				cat.angle = lookAngle

				cat.x = prevCat.x + math.cos(cat.angle + math.pi) * 135
				cat.y = prevCat.y + math.sin(cat.angle + math.pi) * 135

				segs[i-1].x = (cat.x + prevCat.x) * .5
				segs[i-1].y = (cat.y + prevCat.y) * .5
				segs[i-1].angle = lookAngle
			end

			-- re-adjust angles for better looking
			for i=3,#segs-2,2 do
				local angle1 = segs[i-1].angle
				local angle2 = segs[i+1].angle
				if angle1 < math.pi * .5 and angle2 > math.pi * 1.5 then
					angle1 = angle1 + math.pi * 2
				elseif angle1 > math.pi * 1.5 and angle2 < math.pi * 0.5 then
					angle2 = angle2 + math.pi * 2
				end
				segs[i].angle = (angle1 + angle2) * .5
			end
		end

		if caterpillar.slowingDown then
			local targetView = segs[1].x - 512
			viewX = viewX + (targetView - viewX) * time
		end

	elseif fightState == 3 then

		-- in fightState 3 the caterpillar's body parts starts falling off
		local lowestSegment = #segs
		for i=2,#segs do

			if segs[i].fallTimer > 0 then
				segs[i].fallTimer = segs[i].fallTimer - time
				if segs[i].fallTimer <= 0 then
					caterpillar.headShakeTimer = .25
					if i < lowestSegment then
						lowestSegment = i
					end
				end
			else
				segs[i].vy = segs[i].vy + time * 1000
				segs[i].x = segs[i].x + segs[i].vx * time
				segs[i].y = segs[i].y + segs[i].vy * time
				segs[i].angle = segs[i].angle + segs[i].vy * time * .01
			end
		end

		if lowestSegment <= 2 then
			-- all body parts are flying off, so continously shake the head and move it out from the screen
			caterpillar.isRetreating = true
		end

		if caterpillar.isRetreating and not caterpillar.hasRetreated then
			caterpillar.headShakeTimer = .25
			segs[1].y = segs[1].y - time * 75

			if segs[1].y < -400 then
				caterpillar.hasRetreated = true
				daisy.stopMusic("quakeLoop.wav")

				-- only keep the head segment
				caterpillar.segments = { segs[1] }
				segs = caterpillar.segments
				--print("Segments " .. #segs)

				addMutation(mutationTypes[5], segs[1], 1)
			end
		elseif caterpillar.hasRetreated then
			segs[1].y = segs[1].y + (300 - segs[1].y) * time * 4
			segs[1].angle = math.pi * .5

			if segs[1].y >= 290 then
				fightState = 4
				--daisy.playMusic("darkForce.mp3", 0.75, true)
			end
		end

	end
end

--    GGGGGG
--  GG      GG
--  GG
--  GG    GGGG
--  GG      GG
--    GGGGGG   U I

local function updatePlayerGuiInput(state)

	if state.done then
		return
	end

	if state.state == 0 then
		state.selectionIndex = 1
		if not state.up and state.prevUp then
			state.state = 1
			daisy.playSound("bclick.wav")
		elseif not state.down and state.prevDown then
			state.state = 2
			daisy.playSound("bclick.wav")
		elseif not state.right and state.prevRight and state.mana[2] >= lowestPossibleMana then
			state.state = 3
			daisy.playSound("bclick.wav")
			-- locate the first possible spell
			for i=1, #(state.spells) do
				if state.spells[i].manaCost <= state.mana[2] then
					state.selectionIndex = i
					break
				end
			end
		elseif not state.left and state.prevLeft then
			if #state.items > 0 then
				state.state = 4
				daisy.playSound("bclick.wav")
			end
		end
	elseif state.state == 1 then
		if not state.up and state.prevUp then
			state.done = true
			state.selectionAction = 1
			daisy.playSound("bclick.wav")
		elseif not state.down and state.prevDown then
			state.state = 0
			daisy.playSound("bclick.wav")
		end

		if state.left and not state.prevLeft then
			state.selectionIndex = state.selectionIndex - 1
			if state.selectionIndex == 0 then
				state.selectionIndex = #caterpillar.mutations
			end
			daisy.playSound("bclick.wav")
		elseif state.right and not state.prevRight then
			state.selectionIndex = state.selectionIndex + 1
			if state.selectionIndex > #caterpillar.mutations then
				state.selectionIndex = 1
			end
			daisy.playSound("bclick.wav")
		end

	elseif state.state == 2 then
		if not state.up and state.prevUp then
			state.state = 0
			daisy.playSound("bclick.wav")
		elseif not state.down and state.prevDown then
			state.done = true
			state.selectionAction = 2
			daisy.playSound("bclick.wav")
		end
	elseif state.state == 3 then
		if not state.left and state.prevLeft then
			state.state = 0
			daisy.playSound("bclick.wav")
		elseif not state.right and state.prevRight then
			state.subIndex = 1
			state.state = 5
			daisy.playSound("bclick.wav")
		end

		if state.up and not state.prevUp then
			repeat
				state.selectionIndex = state.selectionIndex - 1
				if state.selectionIndex == 0 then
					state.selectionIndex = #state.spells
				end
			until state.spells[state.selectionIndex].manaCost <= state.mana[2]
			daisy.playSound("bclick.wav")
		elseif state.down and not state.prevDown then
			repeat
				state.selectionIndex = state.selectionIndex + 1
				if state.selectionIndex > #state.spells then
					state.selectionIndex = 1
				end
			until state.spells[state.selectionIndex].manaCost <= state.mana[2]
			daisy.playSound("bclick.wav")
		end
	elseif state.state == 4 then
		if not state.left and state.prevLeft then

			if state.items[state.selectionIndex].spell then
				state.subIndex = 1
				state.state = 6
				daisy.playSound("bclick.wav")
			end

		elseif not state.right and state.prevRight then
			state.state = 0
			daisy.playSound("bclick.wav")
		end

		if state.up and not state.prevUp then
			state.selectionIndex = state.selectionIndex - 1
			if state.selectionIndex == 0 then
				state.selectionIndex = 1
			end
			daisy.playSound("bclick.wav")
		elseif state.down and not state.prevDown then
			state.selectionIndex = state.selectionIndex + 1
			if state.selectionIndex > #state.items then
				state.selectionIndex = #state.items
			end
			daisy.playSound("bclick.wav")
		end
	elseif state.state == 5 then
		if not state.down and state.prevDown then
			state.state = 3 -- back to spell list
			daisy.playSound("bclick.wav")
		elseif not state.up and state.prevUp then

			-- verify targets
			local performed = false
			local spell = state.spells[state.selectionIndex]

			if spell.type == 1 then
				performed = spell.testMethod(playerCharacters[state.subIndex])
			elseif spell.type == 2 then
				performed = spell.testMethod(playerCharacters[1]) or spell.testMethod(playerCharacters[2])
			elseif spell.type == 3 then
				performed = spell.testMethod(caterpillar.mutations[state.subIndex])
			elseif spell.type == 4 then
				for i=1,#caterpillar.mutations do
					if spell.testMethod(caterpillar.mutations[i]) then
						performed = true
						break
					end
				end
			end

			if performed then
				state.done = true
				state.selectionAction = 3
			end
			daisy.playSound("bclick.wav")

		elseif state.left and not state.prevLeft then
			local spell = state.spells[state.selectionIndex]

			if spell.type == 1 then
				if state.subIndex == 1 then
					state.subIndex = 2
				else
					state.subIndex = 1
				end
			elseif spell.type == 3 then
				state.subIndex = state.subIndex - 1
				if state.subIndex == 0 then
					state.subIndex = #caterpillar.mutations
				end
			end
			daisy.playSound("bclick.wav")
		elseif state.right and not state.prevRight then
			local spell = state.spells[state.selectionIndex]

			if spell.type == 1 then
				if state.subIndex == 1 then
					state.subIndex = 2
				else
					state.subIndex = 1
				end
			elseif spell.type == 3 then
				state.subIndex = state.subIndex + 1
				if state.subIndex > #caterpillar.mutations then
					state.subIndex = 1
				end
			end
			daisy.playSound("bclick.wav")
		end
	elseif state.state == 6 then
		if not state.down and state.prevDown then
			state.state = 4 -- back to item list
			daisy.playSound("bclick.wav")
		elseif not state.up and state.prevUp then

			-- verify targets
			local performed = false
			local spell = state.items[state.selectionIndex].spell

			if spell.type == 1 then
				performed = spell.testMethod(playerCharacters[state.subIndex])
			elseif spell.type == 2 then
				performed = spell.testMethod(playerCharacters[1]) or spell.testMethod(playerCharacters[2])
			elseif spell.type == 3 then
				performed = spell.testMethod(caterpillar.mutations[state.subIndex])
			elseif spell.type == 4 then
				for i=1,#caterpillar.mutations do
					if spell.testMethod(caterpillar.mutations[i]) then
						performed = true
						break
					end
				end
			end

			if performed then
				state.done = true
				state.selectionAction = 4
			end
			daisy.playSound("bclick.wav")

		elseif state.left and not state.prevLeft then
			local spell = state.items[state.selectionIndex].spell

			if spell.type == 1 then
				if state.subIndex == 1 then
					state.subIndex = 2
				else
					state.subIndex = 1
				end
			elseif spell.type == 3 then
				state.subIndex = state.subIndex - 1
				if state.subIndex == 0 then
					state.subIndex = #caterpillar.mutations
				end
			end
			daisy.playSound("bclick.wav")
		elseif state.right and not state.prevRight then
			local spell = state.items[state.selectionIndex].spell

			if spell.type == 1 then
				if state.subIndex == 1 then
					state.subIndex = 2
				else
					state.subIndex = 1
				end
			elseif spell.type == 3 then
				state.subIndex = state.subIndex + 1
				if state.subIndex > #caterpillar.mutations then
					state.subIndex = 1
				end
			end
			daisy.playSound("bclick.wav")
		end
	end

	state.prevLeft = state.left
	state.prevUp = state.up
	state.prevRight = state.right
	state.prevDown = state.down
end

local function renderPlayerGuiInput(state, x)

	if state.done then
		return
	end

	local y = PY - 100
	if state.state == 0 then

		-- attack square
		if state.prevUp then
			daisy.renderSpriteState(guiSprites.attack, x, y - 40, 1)
			daisy.renderText("Attack", x, y - 82, 1, "gameFont.fnt")
		else
			daisy.renderSpriteState(guiSprites.attack, x, y - 40, 0.75)
		end
		-- item square
		if #state.items > 0 then
			if state.prevLeft then
				daisy.renderSpriteState(guiSprites.item, x - 40, y, 1)
				daisy.renderText("Item", x - 70, y - 10, 2, "gameFont.fnt")
			else
				daisy.renderSpriteState(guiSprites.item, x - 40, y, 0.75)
			end
		end
		-- magic square
		if state.mana[2] >= lowestPossibleMana then
			if state.prevRight then
				daisy.renderSpriteState(guiSprites.spell, x + 40, y, 1)
				daisy.renderText("Magic", x + 70, y - 10, 0, "gameFont.fnt")
			else
				daisy.renderSpriteState(guiSprites.spell, x + 40, y, 0.75)
			end
		end
		-- defense square
		if state.prevDown then
			daisy.renderSpriteState(guiSprites.defend, x, y + 40, 1)
			daisy.renderText("Defend", x, y + 70, 1, "gameFont.fnt")
		else
			daisy.renderSpriteState(guiSprites.defend, x, y + 40, 0.75)
		end

	elseif state.state == 1 then

		local segment = caterpillar.mutations[state.selectionIndex].seg
		local cx,cy = segment.x - viewX, segment.y

		--daisy.renderRectangle(cx - 50, cy - 50, 100, 100, 128, 255, 0, 0)
		daisy.renderSpriteState(guiSprites.target, cx, cy, 1.2)

		-- confirm square
		if state.prevUp then
			daisy.renderSpriteState(guiSprites.confirm, x, y - 40, 1)
		else
			daisy.renderSpriteState(guiSprites.confirm, x, y - 40, 0.75)
		end
		-- cancel square
		if state.prevDown then
			daisy.renderSpriteState(guiSprites.cancel, x, y + 40, 1)
		else
			daisy.renderSpriteState(guiSprites.cancel, x, y + 40, 0.75)
		end

		daisy.renderText("Select Target", x, y - 8, 1, "gameFont.fnt")

	elseif state.state == 2 then

		-- confirm square
		if state.prevDown then
			daisy.renderSpriteState(guiSprites.confirm, x, y + 40, 1)
		else
			daisy.renderSpriteState(guiSprites.confirm, x, y + 40, 0.75)
		end
		-- cancel square
		if state.prevUp then
			daisy.renderSpriteState(guiSprites.cancel, x, y - 40, 1)
		else
			daisy.renderSpriteState(guiSprites.cancel, x, y - 40, 0.75)
		end

		daisy.renderText("Confirm Defense", x, y - 8, 1, "gameFont.fnt")

	elseif state.state == 3 then

		-- selection
		daisy.renderRectangle(x - 50, y - 10, 100, 20, 128, 0, 0, 0)

		for i=state.selectionIndex-2,state.selectionIndex+2 do
			if i >= 1 and i <= #state.spells then
				local textY = y - 8 + (i - state.selectionIndex) * 20
				local alpha = 255 - math.abs(i - state.selectionIndex) * 64
				local intensity = 255
				if state.spells[i].manaCost > state.mana[2] then
					intensity = 128
				end
				daisy.renderText(state.spells[i].name, x, textY, 1, "gameFont.fnt", alpha, intensity, intensity, intensity)
			end
		end

		-- confirm square
		if state.prevRight then
			daisy.renderSpriteState(guiSprites.confirm, x + 80, y, 1)
		else
			daisy.renderSpriteState(guiSprites.confirm, x + 80, y, 0.75)
		end
		-- cancel square
		if state.prevLeft then
			daisy.renderSpriteState(guiSprites.cancel, x - 80, y, 1)
		else
			daisy.renderSpriteState(guiSprites.cancel, x - 80, y, 0.75)
		end

	elseif state.state == 4 then

		-- selection
		--daisy.renderRectangle(x - 50, y - 10, 100, 20, 128, 0, 0, 0)
		local powerupSprites = powerupSprites

		for i=state.selectionIndex-2,state.selectionIndex+2 do
			if i >= 1 and i <= #state.items then
				local textY = y + (i - state.selectionIndex) * 32
				local alpha = 255 - math.abs(i - state.selectionIndex) * 80
				if i == state.selectionIndex then
					daisy.renderSpriteState(powerupSprites.glow, x, textY, 1, 0, alpha)
				end
				daisy.renderSpriteState(powerupSprites[ state.items[i].sprite ], x, textY, 1, 0, alpha)
				--daisy.renderText(state.items[i], x, textY, 1, "gameFont.fnt", alpha, 255, 255, 255)
			end
		end

		-- confirm square
		if state.prevLeft then
			daisy.renderSpriteState(guiSprites.confirm, x - 40, y, 1)
		else
			daisy.renderSpriteState(guiSprites.confirm, x - 40, y, 0.75)
		end

		-- cancel square
		if state.prevRight then
			daisy.renderSpriteState(guiSprites.cancel, x + 40, y, 1)
		else
			daisy.renderSpriteState(guiSprites.cancel, x + 40, y, 0.75)
		end

	elseif state.state == 5 then

		local spell = state.spells[state.selectionIndex]


		if spell.type == 1 then
			if state.subIndex == 1 then
				--daisy.renderRectangle(PX1 - 50, PY - 50, 100, 100, 128, 255, 0, 0)
				daisy.renderSpriteState(guiSprites.target, PX1, PY, 1)
			else
				--daisy.renderRectangle(PX2 - 50, PY - 50, 100, 100, 128, 255, 0, 0)
				daisy.renderSpriteState(guiSprites.target, PX2, PY, 1)
			end
		elseif spell.type == 2 then
			--daisy.renderRectangle(PX1 - 50, PY - 50, 100, 100, 128, 255, 0, 0)
			--daisy.renderRectangle(PX2 - 50, PY - 50, 100, 100, 128, 255, 0, 0)
			daisy.renderSpriteState(guiSprites.target, PX1, PY, 1)
			daisy.renderSpriteState(guiSprites.target, PX2, PY, 1)
		elseif spell.type == 3 then
			local segment = caterpillar.mutations[state.subIndex].seg
			local cx,cy = segment.x - viewX, segment.y

			--daisy.renderRectangle(cx - 50, cy - 50, 100, 100, 128, 255, 0, 0)
			daisy.renderSpriteState(guiSprites.target, cx, cy, 1.2)
		elseif spell.type == 4 then
			for i=1,#caterpillar.mutations do
				local segment = caterpillar.mutations[i].seg
				local cx,cy = segment.x - viewX, segment.y

				--daisy.renderRectangle(cx - 50, cy - 50, 100, 100, 128, 255, 0, 0)
				daisy.renderSpriteState(guiSprites.target, cx, cy, 1.2)
			end
		end

		-- selection
		daisy.renderRectangle(x - 50, y - 10, 100, 20, 128, 0, 0, 0)
		daisy.renderText(spell.name, x, y - 8, 1, "gameFont.fnt", 255, 255, 255, 255)

		-- confirm square
		if state.prevUp then
			daisy.renderSpriteState(guiSprites.confirm, x, y - 40, 1)
		else
			daisy.renderSpriteState(guiSprites.confirm, x, y - 40, 0.75)
		end
		-- cancel square
		if state.prevDown then
			daisy.renderSpriteState(guiSprites.cancel, x, y + 40, 1)
		else
			daisy.renderSpriteState(guiSprites.cancel, x, y + 40, 0.75)
		end

	elseif state.state == 6 then

		local spell = state.items[state.selectionIndex].spell


		if spell.type == 1 then
			if state.subIndex == 1 then
				daisy.renderSpriteState(guiSprites.target, PX1, PY, 1)
			else
				daisy.renderSpriteState(guiSprites.target, PX2, PY, 1)
			end
		elseif spell.type == 2 then
			daisy.renderSpriteState(guiSprites.target, PX1, PY, 1)
			daisy.renderSpriteState(guiSprites.target, PX2, PY, 1)
		elseif spell.type == 3 then
			local segment = caterpillar.mutations[state.subIndex].seg
			local cx,cy = segment.x - viewX, segment.y
			daisy.renderSpriteState(guiSprites.target, cx, cy, 1.2)
		elseif spell.type == 4 then
			for i=1,#caterpillar.mutations do
				local segment = caterpillar.mutations[i].seg
				local cx,cy = segment.x - viewX, segment.y
				daisy.renderSpriteState(guiSprites.target, cx, cy, 1.2)
			end
		end

		-- selection
		daisy.renderSpriteState(powerupSprites.glow, x, y, 1, 0)
		daisy.renderSpriteState(powerupSprites[ state.items[state.selectionIndex].sprite ], x, y, 1, 0)

		-- confirm square
		if state.prevUp then
			daisy.renderSpriteState(guiSprites.confirm, x, y - 40, 1)
		else
			daisy.renderSpriteState(guiSprites.confirm, x, y - 40, 0.75)
		end
		-- cancel square
		if state.prevDown then
			daisy.renderSpriteState(guiSprites.cancel, x, y + 40, 1)
		else
			daisy.renderSpriteState(guiSprites.cancel, x, y + 40, 0.75)
		end
	end
end

--   SSSSSS    PPPPPP    EEEEEE    LLLLLL    LLLLLL    SSSSSS
-- SS      SS  PPPPPP    EEEEEE    LLLLLL    LLLLLL    SSSSSS
--   SSS       PPPPPP    EEEEEE    LLLLLL    LLLLLL    SSSSSS
--      SSS    PPPPPP    EEEEEE    LLLLLL    LLLLLL    SSSSSS
-- SS      SS  PPPPPP    EEEEEE    LLLLLL    LLLLLL    SSSSSS
--   SSSSSS    PPPPPP    EEEEEE    LLLLLL    LLLLLL    SSSSSS
local function spellTestLak(player)
	return player.health[2] > 0 and player.health[2] < player.health[4]
end
spells.lak.testMethod = spellTestLak

local function spellActionLak(source, player)
	local amount = source.magic * 2 + math.random(source.magic)
	if (player.health[2] + amount) > player.health[4] then
		amount = player.health[4] - player.health[2]
	end
	player.health[2] = player.health[2] + amount
	addTextFloater("+" .. amount, player.screenX, PY)
end
spells.lak.actionMethod = spellActionLak

local function spellTestDuLak(player)
	return player.health[2] > 0 and player.health[2] < player.health[4]
end
spells.dulak.testMethod = spellTestDuLak

local function spellActionDuLak(source, player)
	local amount = source.magic + math.random(source.magic)
	if (player.health[2] + amount) > player.health[4] then
		amount = player.health[4] - player.health[2]
	end
	player.health[2] = player.health[2] + amount
	addTextFloater("+" .. amount, player.screenX, PY)
end
spells.dulak.actionMethod = spellActionDuLak

local function spellTestUppsta(player)
	return player.health[2] <= 0
end
spells.uppsta.testMethod = spellTestUppsta

local function spellActionUppsta(source, player)
	player.health[2] = player.health[4]
	addTextFloater("+" .. player.health[4], player.screenX, PY)
end
spells.uppsta.actionMethod = spellActionUppsta

local function spellTestSkold(player)
	return player.health[2] > 0 and player.isShielded <= 0
end
spells.skold.testMethod = spellTestSkold

local function spellActionSkold(source, player)
	player.isShielded = math.floor(source.magic / 4)
	player.defense = player.defense + player.isShielded
	addTextFloater("DEF UP", player.screenX, PY)
end
spells.skold.actionMethod = spellActionSkold

local function spellTestHast(player)
	return player.health[2] > 0 and player.isHasted <= 0
end
spells.hast.testMethod = spellTestHast

local function spellActionHast(source, player)
	player.isHasted = math.floor(source.magic / 4)
	player.agility = player.agility + player.isHasted
	addTextFloater("AGI UP", player.screenX, PY)
end
spells.hast.actionMethod = spellActionHast

local function spellTestBrann(mut)
	return mut.health > 0
end
spells.brann.testMethod = spellTestBrann

local function spellActionBrann(source, mut)
	local damage = source.magic * 2 + math.random(source.magic) - math.floor(mut.type.defense + mut.type.magic * .5)
	if mut.isDefending then
		damage = 0
	end
	if damage < 0 then
		damage = 0
	else
		mut.health = mut.health - damage
		if mut.health < 0 then
			mut.health = 0
		end
		caterpillar.headShakeTimer = .25
	end
	addTextFloater("-" .. damage, mut.seg.x - viewX, mut.seg.y)
end
spells.brann.actionMethod = spellActionBrann

local function spellTestBlixt(mut)
	return mut.health > 0
end
spells.blixt.testMethod = spellTestBlixt

local function spellActionBlixt(source, mut)
	local damage = source.magic + math.random(source.magic) - math.floor(mut.type.defense + mut.type.magic * .5)
	if mut.isDefending then
		damage = 0
	end
	if damage < 0 then
		damage = 0
	else
		mut.health = mut.health - damage
		if mut.health < 0 then
			mut.health = 0
		end
		caterpillar.headShakeTimer = .25
	end
	addTextFloater("-" .. damage, mut.seg.x - viewX, mut.seg.y)
end
spells.blixt.actionMethod = spellActionBlixt

local function spellTestSyringe(player)
	return player.health[2] > 0 and player.mana[2] < player.mana[4]
end
spells.syringe.testMethod = spellTestSyringe

local function spellActionSyringe(source, player)
	local amount = 40 + math.random(30)
	if (player.mana[2] + amount) > player.mana[4] then
		amount = player.mana[4] - player.mana[2]
	end
	player.mana[2] = player.mana[2] + amount
	addTextFloater("+" .. amount, player.screenX, PY)
end
spells.syringe.actionMethod = spellActionSyringe

local function spellTestDryer(player)
	return dryerCount < 2
end
spells.dryer.testMethod = spellTestDryer

local function spellActionDryer(source, player)
	dryerCount = 2
end
spells.dryer.actionMethod = spellActionDryer

--   AAAAAA      CCCCCC    TTTTTTTT
-- AA      AA  CC      CC     TT
-- AAAAAAAAAA  CC             TT
-- AA      AA  CC             TT
-- AA      AA  CC      CC     TT
-- AA      AA    CCCCCC       TT    IONS
local function startNextRoundAction()
	roundDelays.firstWait = 1
	roundDelays.actionWait = 1
	roundDelays.resultWait = 1
	roundDelays.extraEvent = false

	if #roundActions == 0 then
		-- out of actions, start next round
		roundState = 0

		playerCharacters[1].done = not (playerCharacters[1].health[2] > 0)
		playerCharacters[1].state = 0
		playerCharacters[1].isDefending = false
		playerCharacters[2].done = not (playerCharacters[2].health[2] > 0)
		playerCharacters[2].state = 0
		playerCharacters[2].isDefending = false

		if dryerCount > 0 then
			dryerCount = dryerCount - 1
		end

		activePlayer = 0
		if playerCharacters[1].done then
			activePlayer = 1
		end

		for i,mut in pairs(caterpillar.mutations) do
			mut.isDefending = false
		end

		-- remove killed mutations (there's some bug with table.remove if you remove several items, so keep looping)
		local foundKilled = false
		repeat
			foundKilled = false

			for i,mut in pairs(caterpillar.mutations) do
				if mut.health <= 0 then
					table.remove(caterpillar.mutations, i)
					foundKilled = true
					break
				end
			end
		until not foundKilled

		if #caterpillar.mutations == 0 then
			if fightState == 2 then
				fightState = 3
				--daisy.stopAllMusic()
				daisy.playMusic("quakeLoop.wav", 0.8, true)

				local count = #(caterpillar.segments)
				for i=2,count do
					caterpillar.segments[i].fallTimer = (count - i) * .25
					caterpillar.segments[i].vy = math.random(-200, 0)
				end

			elseif fightState == 4 then
				-- we have won!
				gameOver = true
				isVictory = true
				daisy.stopAllMusic()
			end
		elseif playerCharacters[1].health[2] <= 0 and playerCharacters[2].health[2] <= 0 then
			-- we have lost!
			gameOver = true
			isVictory = false
			daisy.stopAllMusic()
		end

	else

	end
end

local function startPlayerAction(action)

	local stats = playerCharacters[action.source]
	if stats.health[2] <= 0 then
		return false
	end

	if action.action == 1 then
		local mut = caterpillar.mutations[action.target]
		if mut.health <= 0 then
			return false
		end

		daisy.playSound("attack.wav")
		roundDelays.actionWait = 0.60

	elseif action.action == 2 then

	elseif action.action == 3 then

		-- target is spell index, subTarget is actual target
		local spell = stats.spells[action.target]

		if stats.mana[2] < spell.manaCost then
			return false
		end

		local performed = false

		if spell.type == 1 then
			performed = spell.testMethod(playerCharacters[action.subTarget])
		elseif spell.type == 2 then
			performed = spell.testMethod(playerCharacters[1]) or spell.testMethod(playerCharacters[2])
		elseif spell.type == 3 then
			performed = spell.testMethod(caterpillar.mutations[action.subTarget])
		elseif spell.type == 4 then
			for i=1,#caterpillar.mutations do
				if spell.testMethod(caterpillar.mutations[i]) then
					performed = true
					break
				end
			end
		end

		if performed then
			daisy.playSound(spell.sound)
			roundDelays.actionWait = spell.timer
		end

		return performed

	elseif action.action == 4 then

		local item = stats.items[action.target]
		if not item.spell then
			return false
		end

		local spell = item.spell
		if spell.type == 1 then
			performed = spell.testMethod(playerCharacters[action.subTarget])
		elseif spell.type == 2 then
			performed = spell.testMethod(playerCharacters[1]) or spell.testMethod(playerCharacters[2])
		elseif spell.type == 3 then
			performed = spell.testMethod(caterpillar.mutations[action.subTarget])
		elseif spell.type == 4 then
			for i=1,#caterpillar.mutations do
				if spell.testMethod(caterpillar.mutations[i]) then
					performed = true
					break
				end
			end
		end

		if performed then
			daisy.playSound(spell.sound)
			roundDelays.actionWait = spell.timer
		end

		return performed
	end

	return true
end

local function startMonsterAction(action)
	local mut = caterpillar.mutations[action.source]

	if mut.health <= 0 then
		return false
	end

	local actionType = mut.type.actionType
	if actionType == 5 then
		-- the boss can have different attacks
		actionType = action.action
	end

	if actionType == 1 then
		-- attacker
		local target = playerCharacters[action.target]
		if target.health[2] <= 0 then
			return false
		end

		roundDelays.actionWait = 1.5
		daisy.playSound("bossattack.wav")

	elseif actionType == 2 then
		-- healer
		local target = caterpillar.mutations[action.target]
		if target.health <= 0 or target.health >= target.type.startHealth then
			return false
		end
		daisy.playSound("bossheal.wav")
		roundDelays.actionWait = 0.1

	elseif actionType == 3 then
		-- vampire
		local target = playerCharacters[action.target]
		if target.health[2] <= 0 then
			return false
		end
		roundDelays.actionWait = 1.7
		daisy.playSound("vampire.wav")

	elseif actionType == 4 then
		-- clockwork

		if playerCharacters[1].health[2] <= 0 and playerCharacters[2].health[2] <= 0 then
			return false
		end

		roundDelays.actionWait = 4.1
		daisy.playSound("quake.wav")

	elseif actionType == 5 then
		-- defending

	end

	mut.actionTimer = roundDelays.actionWait + 1
	return true
end

local function performPlayerAction(action)

	local stats = playerCharacters[action.source]
	if stats.health[2] <= 0 then
		return
	end

	if action.action == 1 then
		local mut = caterpillar.mutations[action.target]
		local damage = stats.attack * 2 + math.random(stats.attack) - math.floor(mut.type.defense + mut.type.agility * .5)

		if mut.isDefending then
			damage = 0
		end
		if damage < 0 then
			damage = 0
		else
			mut.health = mut.health - damage
			if mut.health < 0 then
				mut.health = 0
			end
			caterpillar.headShakeTimer = .25
		end
		addTextFloater("-" .. damage, mut.seg.x - viewX, mut.seg.y)

	elseif action.action == 2 then

		stats.isDefending = true
		addTextFloater("Defending", stats.screenX, PY)

	elseif action.action == 3 then

		-- target is spell index, subTarget is actual target
		local spell = stats.spells[action.target]

		if spell.type == 1 then
			spell.actionMethod(stats, playerCharacters[action.subTarget])
		elseif spell.type == 2 then
			if spell.testMethod(playerCharacters[1]) then
				spell.actionMethod(stats, playerCharacters[1])
			end
			if spell.testMethod(playerCharacters[2]) then
				spell.actionMethod(stats, playerCharacters[2])
			end
		elseif spell.type == 3 then
			spell.actionMethod(stats, caterpillar.mutations[action.subTarget])
		elseif spell.type == 4 then
			for i=1,#caterpillar.mutations do
				if spell.testMethod(caterpillar.mutations[i]) then
					spell.actionMethod(stats, caterpillar.mutations[i])
				end
			end
		end

		stats.mana[2] = stats.mana[2] - spell.manaCost


	elseif action.action == 4 then

		-- target is item index, subTarget is actual target
		local spell = stats.items[action.target].spell

		if spell.type == 1 then
			spell.actionMethod(stats, playerCharacters[action.subTarget])
		elseif spell.type == 2 then
			if spell.testMethod(playerCharacters[1]) then
				spell.actionMethod(stats, playerCharacters[1])
			end
			if spell.testMethod(playerCharacters[2]) then
				spell.actionMethod(stats, playerCharacters[2])
			end
		elseif spell.type == 3 then
			spell.actionMethod(stats, caterpillar.mutations[action.subTarget])
		elseif spell.type == 4 then
			for i=1,#caterpillar.mutations do
				if spell.testMethod(caterpillar.mutations[i]) then
					spell.actionMethod(stats, caterpillar.mutations[i])
				end
			end
		end

		-- erase the player's item
		table.remove(stats.items, action.target)
	end
end

local function removePlayerShield(target)
	target.defense = target.defense - target.isShielded
	target.isShielded = 0
end

local function removePlayerHaste(target)
	target.agility = target.agility - target.isHasted
	target.isHasted = 0
end

local function playerDied(target)
	removePlayerShield(target)
	removePlayerHaste(target)

	-- check if the player has an oxeye item
	for index,item in pairs(target.items) do
		if item.sprite == 6 then
			-- revive player!
			target.health[2] = math.floor(target.health[4] * .33)
			addTextFloater("OXEYE DAISY +" .. target.health[2], target.screenX, PY + 25)
			daisy.playSound("aah.ogg")

			-- if the player has an action in the current round it should be cancelled
			for actIndex,action in pairs(roundActions) do
				if action.sourceType == 1 and target == playerCharacters[action.source] then
					table.remove(roundActions, actIndex)
					break
				end
			end

			table.remove(target.items, index)
			break
		end
	end
end

local function performMonsterAction(action)
	local mut = caterpillar.mutations[action.source]

	local actionType = mut.type.actionType
	if actionType == 5 then
		-- the boss can have different attacks
		actionType = action.action
	end

	if actionType == 1 then
		-- attacker
		local target = playerCharacters[action.target]

		local damage = mut.type.attack * 2 + math.random(mut.type.attack) - math.floor(target.defense + target.agility * .5)
		if damage < 0 then
			damage = 0
		end
		if target.isDefending then
			damage = 0
		end
		target.health[2] = target.health[2] - damage
		if target.health[2] <= 0 then
			target.health[2] = 0
			playerDied(target)
		end
		addTextFloater("-" .. damage, target.screenX, PY)
	elseif actionType == 2 then
		-- healer
		local target = caterpillar.mutations[action.target]
		local amount = mut.type.magic * 2 + math.random(mut.type.magic)

		if (target.health + amount) > target.type.startHealth then
			amount = target.type.startHealth - target.health
		end
		target.health = target.health + amount
		addTextFloater("+" .. amount, target.seg.x - viewX, target.seg.y)

	elseif actionType == 3 then
		-- vampire
		local target = playerCharacters[action.target]

		local amount = mut.type.magic + math.random(mut.type.magic) - math.floor(target.defense + target.magic * .5)
		if amount < 1 then
			amount = 1
		end
		if target.isDefending then
			amount = 0
		end

		if (target.health[2] - amount) < 0 then
			amount = target.health[2]
		end
		target.health[2] = target.health[2] - amount
		if target.health[2] <= 0 then
			playerDied(target)
		end

		addTextFloater("-" .. amount, target.screenX, PY)

		-- randomize a mutation to receive the health
		local targetMut = caterpillar.mutations[math.random(#caterpillar.mutations)]
		if targetMut.health > 0 then
			if (targetMut.health + amount) > targetMut.type.startHealth then
				amount = targetMut.type.startHealth - targetMut.health
			end

			targetMut.health = targetMut.health + amount
			addTextFloater("+" .. amount, targetMut.seg.x - viewX, targetMut.seg.y)
		end

	elseif actionType == 4 then
		-- clockwork

		for p=1,2 do
			local target = playerCharacters[p]
			if target.health[2] > 0 then

				local damage = mut.type.magic + math.random(mut.type.magic) - math.floor(target.defense + target.magic * .5)
				if damage < math.floor(mut.type.magic * .5) then
					damage = math.floor(mut.type.magic * .5)
				end
				if target.isDefending then
					damage = math.floor(damage * .5)
				end

				target.health[2] = target.health[2] - damage
				if target.health[2] <= 0 then
					target.health[2] = 0
					playerDied(target)
				end
				addTextFloater("-" .. damage, target.screenX, PY)
			end
		end

	elseif actionType == 5 then
		-- defending
		mut.isDefending = true
		addTextFloater("Defending", mut.seg.x - viewX, mut.seg.y)

	end
end

local function checkExtraPlayerEvent(action)
	return false
end

local function checkExtraMonsterEvent(action)

	local mut = caterpillar.mutations[action.source]

	local actionType = mut.type.actionType
	if actionType == 5 then
		-- the boss can have different attacks
		actionType = action.action
	end

	if actionType == 1 then
		if dryerCount > 0 then
			-- dryer attack
			local damage = 60 + math.random(30) - math.floor(mut.type.defense + mut.type.magic * .5)
			if damage > mut.health then
				damage = mut.health
			end
			if damage > 0 then
				mut.health = mut.health - damage
				caterpillar.headShakeTimer = .25
				addTextFloater("-" .. damage, mut.seg.x - viewX, mut.seg.y)
			end
			daisy.playSound("brann.wav")
			return true
		end
	elseif actionType == 2 then
	elseif actionType == 3 then
		if dryerCount > 0 then
			-- dryer attack
			local damage = 60 + math.random(30) - math.floor(mut.type.defense + mut.type.magic * .5)
			if damage > mut.health then
				damage = mut.health
			end
			if damage > 0 then
				mut.health = mut.health - damage
				caterpillar.headShakeTimer = .25
				addTextFloater("-" .. damage, mut.seg.x - viewX, mut.seg.y)
			end
			daisy.playSound("brann.wav")
			return true
		end
	elseif actionType == 4 then
	elseif actionType == 5 then
	end

	return false
end

local function updateRoundAction(time)
	if (fightState ~= 2 and fightState ~= 4) or roundState ~= 1 or #roundActions == 0 then
		return
	end

	if roundDelays.firstWait > 0 then
		roundDelays.firstWait = roundDelays.firstWait - time

		if roundDelays.firstWait <= 0 then
			local successful = false
			if roundActions[1].sourceType == 1 then
				successful = startPlayerAction(roundActions[1])
			else
				successful = startMonsterAction(roundActions[1])
			end
			if not successful then
				table.remove(roundActions, 1)
				startNextRoundAction()
			end
		end

	elseif roundDelays.actionWait > 0 then
		roundDelays.actionWait = roundDelays.actionWait - time

		if roundDelays.actionWait <= 0 then
			if roundActions[1].sourceType == 1 then
				performPlayerAction(roundActions[1])
			else
				performMonsterAction(roundActions[1])
			end
		end
	else
		roundDelays.resultWait = roundDelays.resultWait - time

		if roundDelays.resultWait <= 0 then

			if not roundDelays.extraEvent then
				if roundActions[1].sourceType == 1 then
					roundDelays.extraEvent = checkExtraPlayerEvent(roundActions[1])
				else
					roundDelays.extraEvent = checkExtraMonsterEvent(roundActions[1])
				end

				if roundDelays.extraEvent then
					roundDelays.resultWait = 2
					return
				end
			end


			table.remove(roundActions, 1)
			startNextRoundAction()
		end
	end
end

local function orderRoundActions(ac1, ac2)
	-- this method is supposed to return true if ac1 is before ac2, so
	-- since higher agi values goes first...
	return ac1.agi > ac2.agi
end

local function createRoundActions()
	roundActions = {}

	if gameOver then return end

	-- insert player actions
	for i=1,2 do
		if playerCharacters[i].health[2] > 0 then
			local action = playerCharacters[i].selectionAction
			local target = playerCharacters[i].selectionIndex
			local subTarget = playerCharacters[i].subIndex
			local agi = playerCharacters[i].agility

			if action == 3 then
				agi = agi + playerCharacters[i].spells[target].agiBonus
			elseif action == 2 then
				agi = agi + 100
			elseif action == 4 then
				agi = agi + playerCharacters[i].items[target].spell.agiBonus
			end
			agi = agi + math.random(1,50)

			table.insert(roundActions, {action=action, target=target, subTarget=subTarget, agi=agi, sourceType=1, source=i})
		end
	end

	-- insert monster actions
	for i=1,#caterpillar.mutations do
		local mut = caterpillar.mutations[i]
		local action = mut.type.actionType
		local subTarget = 0
		local agi = mut.type.agility + math.random(1,50)
		local target = 0

		if action == 5 then
			action = math.random(1,4)
		end

		if action == 1 or action == 3 then
			target = math.random(1,2)
			if playerCharacters[2].health[2] <= 0 then
				target = 1
			elseif playerCharacters[1].health[2] <= 0 then
				target = 2
			end

		elseif action == 2 then
			-- healer
			local lastLeast = 1000
			for j,other in pairs(caterpillar.mutations) do
				if other.health < other.type.startHealth and other.health < lastLeast then
					target = j
					lastLeast = other.health
				end
			end

		elseif action == 4 then
			-- earthquake, hits all
			target = 1
		end

		if target <= 0 and mut.type.actionType == 5 then
			-- if the boss does a failed action, it will perform a defence action instead
			agi = agi + 100
			target = 1
			action = 5
		end

		if target > 0 then
			table.insert(roundActions, {action=action, target=target, subTarget=subTarget, agi=agi, sourceType=2, source=i})
		end
	end

	shellsort(roundActions, orderRoundActions, #roundActions)
	startNextRoundAction()
end

local function bossFightRun(time)

	if time > .06 then
		time = .06
	end

	if introTimer > 0 then
		introTimer = introTimer - time
		return
	end

	if gameOver and extroTimer > 0 then
		extroTimer = extroTimer - time
	end

	updateSegments(time)
	updatePlayerPowers(time)
	updateTextFloaters(time)
	updateplayerCharacters(time)
	
	cornAnimTimer = cornAnimTimer + time
	playerCharacters[2].aliveSpriteOffsets[1].angle = math.sin((cornAnimTimer + .5) * math.pi * .3) * math.pi * -.10
	playerCharacters[2].aliveSpriteOffsets[2].angle = (math.sin(cornAnimTimer * math.pi * .3) + 1) * math.pi * -.10
	playerCharacters[2].aliveSpriteOffsets[3].angle = math.sin((cornAnimTimer + .5) * math.pi * .3) * math.pi * .10
	playerCharacters[2].aliveSpriteOffsets[4].angle = (math.sin(cornAnimTimer * math.pi * .3) + 1) * math.pi * .10
	
	playerCharacters[1].aliveSpriteOffsets[6].x = math.max(0, math.sin(cornAnimTimer * math.pi * 1.7) * 1)
	playerCharacters[1].aliveSpriteOffsets[6].y = -27 + math.max(0, math.sin(cornAnimTimer * math.pi * 1.7) * 1)
	playerCharacters[1].aliveSpriteOffsets[6].angle = math.max(0, math.sin(cornAnimTimer * math.pi * 1.7) * math.pi * .01)

	if singlePlayer then
		if activePlayer == 0 then
			playerCharacters[2].left,playerCharacters[2].right,playerCharacters[2].up,playerCharacters[2].down = false, false, false, false
			playerCharacters[1].left,playerCharacters[1].right,playerCharacters[1].up,playerCharacters[1].down = daisy.isKeyPressed(37), daisy.isKeyPressed(39), daisy.isKeyPressed(38), daisy.isKeyPressed(40)
		else
			playerCharacters[1].left,playerCharacters[1].right,playerCharacters[1].up,playerCharacters[1].down = false, false, false, false
			playerCharacters[2].left,playerCharacters[2].right,playerCharacters[2].up,playerCharacters[2].down = daisy.isKeyPressed(37), daisy.isKeyPressed(39), daisy.isKeyPressed(38), daisy.isKeyPressed(40)
		end
	else
		playerCharacters[1].left,playerCharacters[1].right,playerCharacters[1].up,playerCharacters[1].down = daisy.isKeyPressed(65), daisy.isKeyPressed(68), daisy.isKeyPressed(87), daisy.isKeyPressed(83)
		playerCharacters[2].left,playerCharacters[2].right,playerCharacters[2].up,playerCharacters[2].down = daisy.isKeyPressed(37), daisy.isKeyPressed(39), daisy.isKeyPressed(38), daisy.isKeyPressed(40)
	end

	if fightState == 2 or fightState == 4 then
		if roundState == 0 then
			updatePlayerGuiInput(playerCharacters[1])
			updatePlayerGuiInput(playerCharacters[2])

			if singlePlayer and activePlayer == 0 and playerCharacters[1].done then
				activePlayer = 1
			end

			if playerCharacters[1].done and playerCharacters[2].done then
				roundState = 1
				createRoundActions()
			end
		elseif roundState == 1 then
			updateRoundAction(time)
		end
	end

end

-- RRRRRRRR
-- RR      RR
-- RRRRRRRR
-- RR      RR
-- RR      RR
-- RR      RR E N D E R
local function renderMutation(mut, intensity)
	if mut.timer <= 0 then
		local i = mut.segIndex
		local bodyOff = math.cos(caterpillar.bodyRotation + i * math.pi * .2) * 3
		local xOff = math.cos(mut.seg.angle) * bodyOff
		local yOff = math.sin(mut.seg.angle) * bodyOff

		--crab
		if(mut.type.actionType == 1 ) then

			local clawAngle = (mut.randValue + 0.1) * math.pi +  math.sin(mut.animTimer)*0.06*math.pi
			local clawing = math.max(0,math.sin(mut.animTimer*15)*0.15*math.pi)

			daisy.renderSpriteState(partSprites[1], mut.seg.x - viewX + xOff + math.cos(clawAngle)*8, mut.seg.y + yOff+ math.sin(clawAngle)*8, 1.0, clawAngle, 255, intensity, intensity, intensity)
			daisy.renderSpriteState(partSprites[3], mut.seg.x - viewX + xOff + math.cos(clawAngle+0.06*math.pi)*49  , mut.seg.y + yOff+ math.sin(clawAngle+0.06*math.pi)*49, 1.0, clawAngle+0.0*math.pi+clawing, 255, intensity, intensity, intensity)
			daisy.renderSpriteState(partSprites[2], mut.seg.x - viewX + xOff + math.cos(clawAngle)*37  , mut.seg.y + yOff+ math.sin(clawAngle)*37, 1.0, clawAngle, 255, intensity, intensity, intensity)


		--healer
		elseif(mut.type.actionType == 2 ) then
			local eyeAngle = mut.seg.angle + math.pi
			local eyeScale = 1

			if math.sin(mut.animTimer * math.pi) > 0 then
				local fraction = mut.animTimer - math.floor(mut.animTimer)
				if fraction < .8 then
					eyeAngle = eyeAngle + math.sin(fraction / .8 * math.pi * 2) * math.pi * .15
					eyeScale = 1.0 + math.sin(((fraction - .5) / .3) * math.pi) * .05
				end
				if fraction < .5 then
					eyeScale = 1.0 + math.sin((fraction / .5) * math.pi) * .10
				end
			end

			daisy.renderSpriteState(partSprites[4], mut.seg.x - viewX + xOff, mut.seg.y + yOff, eyeScale, eyeAngle, 255, intensity, intensity, intensity)
		--vampire
		elseif(mut.type.actionType == 3 ) then

			local jawing = math.max(0,math.sin(mut.animTimer*4)*20)
			local jawX = math.cos(mut.seg.angle - math.pi) * jawing
			local jawY = math.sin(mut.seg.angle - math.pi) * jawing

			daisy.renderSpriteState(partSprites[5], mut.seg.x - viewX + xOff, mut.seg.y + yOff, 0.9, mut.seg.angle - math.pi * .5, 255, intensity, intensity, intensity)
			daisy.renderSpriteState(partSprites[6], mut.seg.x - viewX + xOff - jawX, mut.seg.y + yOff-jawY, 0.9, mut.seg.angle - math.pi * .5, 255, intensity, intensity, intensity)
		--gears
		elseif(mut.type.actionType == 4 ) then
			daisy.renderSpriteState(partSprites[7], mut.seg.x - viewX + xOff-21, mut.seg.y + yOff-4, 0.6, mut.animTimer*1.6, 255, intensity, intensity, intensity)
			daisy.renderSpriteState(partSprites[7], mut.seg.x - viewX + xOff+3, mut.seg.y + yOff-20, 0.8, mut.animTimer*1.3, 255, intensity, intensity, intensity)
			daisy.renderSpriteState(partSprites[7], mut.seg.x - viewX + xOff+6, mut.seg.y + yOff+14, 1.0, mut.animTimer, 255, intensity, intensity, intensity)
		else
			daisy.renderSpriteState(circleSprite, mut.seg.x - viewX + xOff, mut.seg.y + yOff, 1.0, 0, 255, mut.type.r, mut.type.g, mut.type.b)
		end
	end
end

local function bossFightRender()

	if gameOver and extroTimer <= 0 then

		if isVictory then
			daisy.renderSpriteState(victorySprite, 0, 0, 2)
			
			daisy.renderText("Victory!", 512, 300, 1, "gameFont.fnt")
			daisy.renderText("You have defeated the tower guardian and Fay is now safe to return to her home!", 300, 330)
			daisy.renderText("Corn's plan to keep her locked up in his fantasy world has failed, but there's", 300, 345)
			daisy.renderText("no grudge in Fay's mind, so maybe friendship can prevail without the kidnapping", 300, 360)
			daisy.renderText("aspect?", 300, 375)

			daisy.renderText("This game was made for the TIGSource Procedural Generation competition in May", 300, 400)
			daisy.renderText("2008. We hope you enjoyed it and thank you for playing! - Oxeye Game Studio", 300, 415)
			daisy.renderText("http://www.oxeyegames.com  -  http://www.tigsource.com", 512, 440, 1)
		else
			daisy.renderRectangle(0, 0, 1024, 768, 255, 48, 20, 76)
			
			daisy.renderText("Game Over!", 512, 300, 1, "gameFont.fnt")
			daisy.renderText("You have been defeated by the tower guardian! Corn's plan to keep Fay locked up in ", 300, 330)
			daisy.renderText("his world of fantasy has back-fired and they are now both dead. This is a sad day.", 300, 345)

			daisy.renderText("Press Shift+R to restart the game and try again!", 512, 370, 1)
		end
		return
	end

	daisy.renderSpriteState(backgroundSprite, 0, 0, 2)

	--daisy.renderRectangle(0, 0, 1024, 768, 255, 48, 20, 76)
	--daisy.renderRectangle(0, 468, 1024, 300, 128, 96, 40, 152)
	--daisy.renderRectangle(0, 568, 1024, 200, 128, 96, 40, 152)

	local segs = caterpillar.segments
	for i=#segs,1,-2 do

		local intensity = 128 + ((#segs - i) / #segs) * 127
		if i > 1 then
			local cat = segs[i-1]
			local scale = 1 -- 0.5 + (#segs - i) * .02
			if scale > 1 then
				scale = 1
			end
			local wingOff = math.cos(caterpillar.wingRotation + i * 10) * .2
			local xOff = math.cos(cat.angle + math.pi * .5) * 30
			local yOff = math.sin(cat.angle + math.pi * .5) * 30
			daisy.renderSpriteState(catSprites[4], cat.x - xOff - viewX, cat.y - yOff, scale, cat.angle + wingOff, 255, intensity, intensity, intensity)
			daisy.renderSpriteState(catSprites[5], cat.x + xOff - viewX, cat.y + yOff, scale, cat.angle - wingOff, 255, intensity, intensity, intensity)
			daisy.renderSpriteState(catSprites[cat.type], cat.x - viewX, cat.y, scale, cat.angle, 255, intensity, intensity, intensity)
		end

		local cat = segs[i]
		if i > 1 then
			local bodyOff = math.cos(caterpillar.bodyRotation + i * math.pi * .2) * 3
			local xOff = math.cos(cat.angle) * bodyOff
			local yOff = math.sin(cat.angle) * bodyOff
			if cat.mutated then
				daisy.renderSpriteState(catSprites[7], cat.x + xOff - viewX, cat.y + yOff, 1, cat.angle, 255, intensity, intensity, intensity)

				for index,mut in pairs(caterpillar.mutations) do
					if mut.seg == cat then
						renderMutation(mut, intensity)
					end
				end
			else
				daisy.renderSpriteState(catSprites[cat.type], cat.x + xOff - viewX, cat.y + yOff, 1, cat.angle, 255, intensity, intensity, intensity)
			end
		else
			if caterpillar.hasRetreated then

				local newAngle = cat.angle + math.cos(caterpillar.bodyRotation * .2) * math.pi * .05
				intensity = 64
				if #caterpillar.mutations > 0 then
					intensity = 64 + math.floor((caterpillar.mutations[1].health / caterpillar.mutations[1].type.startHealth) * 191)
				end
				if caterpillar.headShakeTimer > 0 then
					daisy.renderSpriteState(catSprites[cat.type], cat.x - viewX + math.random(-7,7), cat.y + math.random(-7,7), 2.5, newAngle, 255, 255, intensity, intensity)
				else
					daisy.renderSpriteState(catSprites[cat.type], cat.x - viewX, cat.y, 2.5, newAngle, 255, 255, intensity, intensity)
				end

			elseif caterpillar.headShakeTimer > 0 then
				daisy.renderSpriteState(catSprites[cat.type], cat.x - viewX + math.random(-7,7), cat.y + math.random(-7,7), 1.1, cat.angle, 255, intensity, intensity, intensity)
			else
				daisy.renderSpriteState(catSprites[cat.type], cat.x - viewX, cat.y, 1.1, cat.angle, 255, intensity, intensity, intensity)
			end
		end
	end

	daisy.renderSpriteState(bottomSprite, 0, 600, 2)
	--daisy.renderRectangle(0, 668, 1024, 100, 255, 96, 40, 152)

	renderplayerCharacters()
	renderPlayerPowers()
	renderTextFloaters()

	if (fightState == 2 or fightState == 4) and roundState == 0 then
		if singlePlayer then
			if activePlayer == 0 then
				renderPlayerGuiInput(playerCharacters[1], PX1)
			else
				renderPlayerGuiInput(playerCharacters[2], PX2)
			end
		else
			renderPlayerGuiInput(playerCharacters[1], PX1)
			renderPlayerGuiInput(playerCharacters[2], PX2)
		end
	end

	if introTimer > 0 then
		local alpha = (introTimer / 4) * 255
		if alpha > 255 then
			alpha = 255
		end
		daisy.renderRectangle(0, 0, 1024, 768, alpha, 48, 20, 76)
	elseif gameOver then
		local alpha = ((8 - extroTimer) / 4) * 255
		if alpha > 255 then
			alpha = 255
		end
		daisy.renderRectangle(0, 0, 1024, 768, alpha, 48, 20, 76)
	end

end


function registerBossFightState()
	hook.add("frameUpdate", bossFightRun)
	hook.add("frameRender", bossFightRender)

	-- rand bug
	local value = math.random(300,724)

	catSprites[1] = daisy.createSpriteState("bossHead", "monsters.dat")
	catSprites[2] = daisy.createSpriteState("bossBody", "monsters.dat")
	catSprites[3] = daisy.createSpriteState("bossLink", "monsters.dat")
	catSprites[4] = daisy.createSpriteState("bossLeftWing", "monsters.dat")
	catSprites[5] = daisy.createSpriteState("bossRightWing", "monsters.dat")
	catSprites[6] = daisy.createSpriteState("bossTail", "monsters.dat")
	catSprites[7] = daisy.createSpriteState("bossBodyMutated", "monsters.dat")

	circleSprite = daisy.createSpriteState("Circle", "itt.dat")
	backgroundSprite = daisy.createSpriteState("BossBackground", "gfx.dat")
	victorySprite = daisy.createSpriteState("BackgroundWin", "gfx.dat")
	bottomSprite = daisy.createSpriteState("BossBottom", "gfx.dat")
	--blockSprite = daisy.createSpriteState("Square", "itt.dat")

	partSprites[1] = daisy.createSpriteState("bossPartClawA", "monsters.dat")
	partSprites[2] = daisy.createSpriteState("bossPartClawB", "monsters.dat")
	partSprites[3] = daisy.createSpriteState("bossPartClawC", "monsters.dat")
	partSprites[4] = daisy.createSpriteState("bossPartEyes", "monsters.dat")
	partSprites[5] = daisy.createSpriteState("bossPartVampireA", "monsters.dat")
	partSprites[6] = daisy.createSpriteState("bossPartVampireB", "monsters.dat")
	partSprites[7] = daisy.createSpriteState("bossPartGear", "monsters.dat")

	guiSprites.attack = daisy.createSpriteState("guiAttack", "players.dat")
	guiSprites.defend = daisy.createSpriteState("guiDefend", "players.dat")
	guiSprites.item = daisy.createSpriteState("guiItem", "players.dat")
	guiSprites.spell = daisy.createSpriteState("guiSpell", "players.dat")
	guiSprites.confirm = daisy.createSpriteState("guiConfirm", "players.dat")
	guiSprites.cancel = daisy.createSpriteState("guiCancel", "players.dat")
	guiSprites.target = daisy.createSpriteState("guiTarget", "players.dat")
	
	playerCharacters[1].deathSprite = daisy.createSpriteState("FayTomb", "tiles.dat")
	playerCharacters[1].aliveSprites[1] = daisy.createSpriteState("fayRLegFront", "players.dat")
	playerCharacters[1].aliveSprites[2] = daisy.createSpriteState("fayLLegFront", "players.dat")
	playerCharacters[1].aliveSprites[3] = daisy.createSpriteState("fayRArmFront", "players.dat")
	playerCharacters[1].aliveSprites[4] = daisy.createSpriteState("fayBodyFront", "players.dat")
	playerCharacters[1].aliveSprites[5] = daisy.createSpriteState("fayLArmFront", "players.dat")
	playerCharacters[1].aliveSprites[6] = daisy.createSpriteState("fayHeadFront", "players.dat")
	playerCharacters[2].deathSprite = daisy.createSpriteState("CornTomb", "tiles.dat")
	playerCharacters[2].aliveSprites[1] = daisy.createSpriteState("cornLWingFront", "players.dat")
	playerCharacters[2].aliveSprites[2] = daisy.createSpriteState("cornLWingFront", "players.dat")
	playerCharacters[2].aliveSprites[3] = daisy.createSpriteState("cornRWingFront", "players.dat")
	playerCharacters[2].aliveSprites[4] = daisy.createSpriteState("cornRWingFront", "players.dat")
	playerCharacters[2].aliveSprites[5] = daisy.createSpriteState("cornRLegFront", "players.dat")
	playerCharacters[2].aliveSprites[6] = daisy.createSpriteState("cornLLegFront", "players.dat")
	playerCharacters[2].aliveSprites[7] = daisy.createSpriteState("cornRArmFront", "players.dat")
	playerCharacters[2].aliveSprites[8] = daisy.createSpriteState("cornLArmFront", "players.dat")
	playerCharacters[2].aliveSprites[9] = daisy.createSpriteState("cornBodyFront", "players.dat")
	playerCharacters[2].aliveSprites[10] = daisy.createSpriteState("cornHeadFront", "players.dat")

	addSegment(1)
	for i=1,30 do
		addSegment(3)
		addSegment(2)
	end
	addSegment(3)
	addSegment(6)

	--caterpillar[1].x = 400
	--caterpillar[1].y = 300
	--caterpillar[1].angle = math.pi * 1.5

	--daisy.setMouseVisible(false)
	daisy.stopAllMusic()
end

function unregisterBossFightState()
	hook.remove("frameUpdate", bossFightRun)
	hook.remove("frameRender", bossFightRender)
end
